import gzip
import logging
import os
from pathlib import Path
import sqlite3
from typing import Optional

from . import config
from .strava import StravaWeb
from .sync import database


def find_gpx(d: Path, i: int) -> Optional[Path]:
    for suffix in [".gpx", ".gpx.gz"]:
        p = Path(d, str(i) + suffix)
        if p.exists():
            return p

    return None


def link_backup_activities(
        db: sqlite3.Connection,
        dir_activities: Path, dir_activities_backup: Path) -> None:
    for activity in db.execute("SELECT id, upload_id FROM activity WHERE upload_id IS NOT NULL"):
        activity_id = int(activity['id'])
        upload_id = int(activity['upload_id'])

        if find_gpx(dir_activities, activity_id):
            continue

        backup = find_gpx(dir_activities_backup, activity_id) or find_gpx(dir_activities_backup, upload_id)
        if backup:
            link = Path(dir_activities, str(activity_id) + "".join(backup.suffixes))
            if hasattr(backup, 'hardlink_to'):
                link.hardlink_to(backup)  # type: ignore [attr-defined]
            else:
                os.link(backup, link)  # python 3.9 compat


def download_gpx(strava: StravaWeb, activity_id: int, path: Path) -> None:
    gpx = strava.get_gpx(activity_id)
    filename = Path(path, str(activity_id) + ".gpx.gz")
    tmpfilename = Path(path, str(activity_id) + ".gpx.gz.tmp")
    with gzip.open(tmpfilename, "wb") as f:
        f.write(gpx)
    tmpfilename.replace(filename)


def download_activities(db: sqlite3.Connection, strava: StravaWeb, dir_activities: Path) -> None:
    new = 0

    for activity in db.execute("SELECT id FROM activity WHERE upload_id IS NOT NULL AND has_location_data"):
        activity_id = int(activity['id'])
        if find_gpx(dir_activities, activity_id):
            continue

        logging.debug(f"downloading gpx for activity {activity_id}")
        download_gpx(strava=strava, activity_id=activity_id, path=dir_activities)
        new += 1

    logging.info(f"downloaded gpx for {new} new activities")


def sync(config: config.GpxConfig, strava: StravaWeb):
    config.dir_activities.mkdir(parents=True, exist_ok=True)

    with database(config) as db:
        if config.dir_activities_backup:
            link_backup_activities(
                db=db,
                dir_activities=config.dir_activities,
                dir_activities_backup=config.dir_activities_backup)

        download_activities(db=db, strava=strava, dir_activities=config.dir_activities)



#!/usr/bin/env python3

from __future__ import annotations

from pathlib import Path
import re
from sys import stdin
from sys import stdout


def include(m: re.Match[str]) -> str:
    f = Path(m[1]).read_text()
    return f"<!-- include {m[1]} -->\n{f}<!-- end include -->"


def includes(s: str) -> str:
    return re.sub(r"(?s)<!-- include (\S+) .*? end include -->", include, s)


if __name__ == "__main__":
    stdout.write(includes(stdin.read()))



# type: ignore

import sys

if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata  # isort: skip
else:
    import importlib_metadata  # isort: skip

try:
    __version__ = importlib_metadata.version(__name__)
except importlib_metadata.PackageNotFoundError:
    # package is not installed
    pass



from contextlib import contextmanager
from datetime import datetime
import sqlite3
from typing import Iterator
from typing import Optional

from . import config
from . import sqlite
from .strava import StravaAPI

table_bike = sqlite.Table(
    name='bike',
    columns={
        'id': "TEXT PRIMARY KEY",
        'name': "TEXT",
    },
    from_dict=lambda bike: {
        'id': bike['id'],
        'name': bike['name'],
    },
)

table_activity = sqlite.Table(
    name='activity',
    columns={
        'id': "INTEGER PRIMARY KEY",
        'upload_id': "TEXT",
        'name': "TEXT",
        'start_date': "TEXT",
        'moving_time': "INTEGER",
        'elapsed_time': "INTEGER",
        'distance': "REAL",
        'total_elevation_gain': "REAL",
        'gear_id': "TEXT",
        'type': "TEXT",
        'sport_type': "TEXT",
        'commute': "BOOLEAN",
        'trainer': "BOOLEAN",
        'has_location_data': "BOOLEAN",
    },
    from_dict=lambda activity: {
        'id': activity['id'],
        'upload_id': activity['upload_id'],
        'name': activity['name'],
        'start_date': activity['start_date'],
        'moving_time': activity['moving_time'],
        'elapsed_time': activity['elapsed_time'],
        'distance': activity['distance'],
        'total_elevation_gain': activity['total_elevation_gain'],
        'gear_id': activity['gear_id'],
        'type': activity['type'],
        'sport_type': activity.get('sport_type'),
        'commute': activity['commute'],
        'trainer': activity['trainer'],
        'has_location_data': isinstance(activity['start_latlng'], list) and len(activity['start_latlng']) >= 2,
    },
)

schema = sqlite.Schema(
    # Version of database schema. Bump this whenever the schema changes,
    # tables will be recreated using the stored json data and the new schema.
    version=4,

    tables=[
        table_bike,
        table_activity,
    ],
)


@contextmanager
def database(config: config.DatabaseConfig) -> Iterator[sqlite3.Connection]:
    with sqlite.database(config.strava_sqlite_database, schema) as db:
        yield db


def sync_bikes(strava: StravaAPI, db: sqlite3.Connection) -> None:
    table_bike.upsert(db, strava.get_bikes())


def sync_activities(
    strava: StravaAPI,
    db: sqlite3.Connection,
    before: Optional[datetime] = None,
    incremental: bool = False,
) -> None:
    table_activity.upsert(db, strava.get_activities(before=before), incremental=incremental)


def sync(config: config.SyncConfig, strava: StravaAPI):
    with database(config) as db:
        sync_bikes(strava, db)
        sync_activities(strava, db, incremental=(not config.full))



from datetime import datetime
from datetime import timezone
import sqlite3

import pytest  # type: ignore [import]

from strava_offline import config
from strava_offline.strava import StravaAPI
from strava_offline import sync


def database():
    return sync.database(config.DatabaseConfig(strava_sqlite_database=":memory:"))


def strava(tmp_path):
    token = tmp_path / "token.json"
    token.write_text('{"access_token": "token"}')
    cfg = config.StravaApiConfig(strava_token_filename=token)
    return StravaAPI(config=cfg)


@pytest.mark.vcr
def test_sync_bikes(tmp_path):
    with database() as db:
        # initial sync
        sync.sync_bikes(strava=strava(tmp_path), db=db)

        # check that we have all the bikes we expect
        bikes = [list(row) for row in db.execute(
            "SELECT id, name FROM bike ORDER BY id")]
        assert bikes == [
            ['b123456', 'bike1'],
            ['b234567', 'bike2'],
            ['b345678', 'bike3'],
        ]

        # delete one bike, insert extra bike
        db.execute("DELETE FROM bike WHERE id = 'b123456'")
        db.execute("INSERT INTO bike (id) VALUES ('b456789')")

        # sync again
        sync.sync_bikes(strava=strava(tmp_path), db=db)

        # recheck that we have all the bikes we expect
        bikes = [list(row) for row in db.execute(
            "SELECT id, name FROM bike ORDER BY id")]
        assert bikes == [
            ['b123456', 'bike1'],
            ['b234567', 'bike2'],
            ['b345678', 'bike3'],
        ]


@pytest.mark.vcr
def test_sync_activities(tmp_path):
    before = datetime.fromtimestamp(1610000000, tz=timezone.utc)

    with database() as db:
        # initial sync
        sync.sync_activities(strava=strava(tmp_path), db=db, before=before)

        # check that we have all the activities we expect
        activities = [list(row) for row in db.execute(
            "SELECT id FROM activity ORDER BY id")]
        assert activities == [
            [1234567890],
            [1234567892],
            [1234567894],
            [1234567896],
            [1234567898],
            [1234567900],
            [1234567902],
            [1234567904],
            [1234567906],
            [1234567908],
            [1234567910],
            [1234567912],
            [1234567914],
            [1234567916],
        ]

        # delete newest and oldest activity, insert extra activity
        db.execute("DELETE FROM activity WHERE id = 1234567890 OR id = 1234567916")
        db.execute("INSERT INTO activity (id) VALUES (1234567999)")

        # sync again
        sync.sync_activities(strava=strava(tmp_path), db=db, before=before)

        # recheck that we have all the activities we expect
        activities = [list(row) for row in db.execute(
            "SELECT id FROM activity ORDER BY id")]
        assert activities == [
            [1234567890],
            [1234567892],
            [1234567894],
            [1234567896],
            [1234567898],
            [1234567900],
            [1234567902],
            [1234567904],
            [1234567906],
            [1234567908],
            [1234567910],
            [1234567912],
            [1234567914],
            [1234567916],
        ]

        # delete newest and oldest activity, insert extra activity
        db.execute("DELETE FROM activity WHERE id = 1234567890 OR id = 1234567916")
        db.execute("INSERT INTO activity (id) VALUES (1234567999)")

        # sync again, but only incrementally
        sync.sync_activities(strava=strava(tmp_path), db=db, before=before, incremental=True)

        # recheck that we have all the activities we expect
        activities = [list(row) for row in db.execute(
            "SELECT id FROM activity ORDER BY id")]
        assert activities == [
            # no 1234567890, incremental sync doesn't go that far
            [1234567892],
            [1234567894],
            [1234567896],
            [1234567898],
            [1234567900],
            [1234567902],
            [1234567904],
            [1234567906],
            [1234567908],
            [1234567910],
            [1234567912],
            [1234567914],
            [1234567916],
            [1234567999],  # incremental doesn't delete extra entries
        ]


@pytest.mark.vcr
def test_migration_bikes(tmp_path):
    db_uri = "file:test_migration_bikes?mode=memory&cache=shared"
    cfg = config.DatabaseConfig(strava_sqlite_database=db_uri)
    db_keep_in_memory = sqlite3.connect(db_uri)

    with sync.database(cfg) as db:
        sync.sync_bikes(strava=strava(tmp_path), db=db)
        db.execute("UPDATE bike SET name = 'xxx'")
        db.execute("PRAGMA user_version = 0")
        db.commit()

    with sync.database(cfg) as db:
        bikes = [row['name'] for row in db.execute(
            "SELECT name FROM bike ORDER BY id LIMIT 1")]
        assert bikes == ['bike1']

    db_keep_in_memory.close()


@pytest.mark.vcr
def test_migration_activities(tmp_path):
    db_uri = "file:test_migration_activities?mode=memory&cache=shared"
    cfg = config.DatabaseConfig(strava_sqlite_database=db_uri)
    db_keep_in_memory = sqlite3.connect(db_uri)

    before = datetime.fromtimestamp(1610000000, tz=timezone.utc)

    with sync.database(cfg) as db:
        sync.sync_activities(strava=strava(tmp_path), db=db, before=before)
        db.execute("UPDATE activity SET name = 'xxx'")
        db.execute("PRAGMA user_version = 0")
        db.commit()

    with sync.database(cfg) as db:
        activities = [row['name'] for row in db.execute(
            "SELECT name FROM activity ORDER BY id LIMIT 1")]
        assert activities == ['name1']

    db_keep_in_memory.close()



import os
import pathlib
from sys import stdout
from typing import Callable
from typing import Iterator
from typing import Optional
from typing import Set

import click
import click_config_file  # type: ignore [import]
import platformdirs
import yaml


def yaml_config_option():
    path = os.path.join(platformdirs.user_config_dir(appname=__package__), 'config.yaml')

    def provider(file_path, _cmd_name):
        if os.path.isfile(file_path):
            with open(file_path) as f:
                return yaml.safe_load(f)
        else:
            return {}

    return click_config_file.configuration_option(
        implicit=False, default=path, show_default=True, provider=provider)


def yaml_config_sample_option(**kwargs):
    def callback(ctx, _param, value) -> None:
        if not value or ctx.resilient_parsing:
            return

        stdout.write(yaml_config_sample(ctx.command, **kwargs))
        ctx.exit()

    return click.option(
        '--config-sample', is_flag=True, is_eager=True, expose_value=False,
        help="Show sample configuration file",
        callback=callback)


def yaml_config_sample(
    command: click.Command,
    sample_get_value: Optional[Callable[[click.Option], Optional[str]]] = None,
    sample_hidden: Set[str] = set()
) -> str:
    sample_hidden = sample_hidden | {'config', 'config_sample'}

    def collect_options(cmd: click.Command, seen: Set[str] = set()) -> Iterator[click.Option]:
        for p in cmd.params:
            if not isinstance(p, click.Option) or not isinstance(p.name, str):
                continue
            if p.hidden or p.name in sample_hidden or p.name in seen:
                continue
            seen.add(p.name)
            yield p

        if isinstance(cmd, click.Group):
            for c in cmd.commands.values():
                yield from collect_options(c, seen=seen)

    def sample_value(opt: click.Option):
        if sample_get_value:
            value = sample_get_value(opt)
            if value is not None:
                return value

        if isinstance(opt.default, (int, str)):
            return opt.default
        elif isinstance(opt.default, pathlib.Path):
            return str(opt.default)
        else:
            return opt.make_metavar()

    def sample_yaml(opt: click.Option):
        sample = ""

        if opt.help:
            sample += "\n".join(f"# {line}" for line in opt.help.splitlines()) + "\n"

        sample += yaml.safe_dump({opt.name: sample_value(opt)}, default_flow_style=False)
        return sample

    return "\n".join(sample_yaml(opt) for opt in collect_options(command))



# type: ignore

import sys

if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata  # isort: skip
else:
    import importlib_metadata  # isort: skip

if __name__ == "__main__":
    _name = importlib_metadata.distribution(__package__).name
    (_script,) = importlib_metadata.entry_points(group='console_scripts', name=_name)
    _main = _script.load()
    sys.exit(_main())



from contextlib import contextmanager
from dataclasses import dataclass
from itertools import chain
import json
import logging
from pathlib import Path
import sqlite3
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Tuple
from typing import Union

from typing_extensions import Protocol


class FromDict(Protocol):
    def __call__(self, data: Mapping[str, Any]) -> Mapping[str, Any]:
        pass


@dataclass(frozen=True)
class Table:
    name: str
    columns: Mapping[str, str]
    from_dict: FromDict

    def _columns(self) -> Iterable[Tuple[str, str]]:
        return chain(self.columns.items(), (('json', "TEXT"),))

    def _from_dict(self, data: Mapping[str, Any]) -> Mapping[str, Any]:
        return {'json': json.dumps(data), **self.from_dict(data)}

    def create(self, db: sqlite3.Connection) -> None:
        columns = ', '.join(f"{name} {type}" for name, type in self._columns())
        db.execute(f"CREATE TABLE IF NOT EXISTS {self.name} ({columns})")

    def migrate(self, db: sqlite3.Connection) -> None:
        for row in db.execute(f"SELECT json FROM {self.name}_old"):
            self.upsert_row(db, self._from_dict(json.loads(row['json'])))
        db.execute(f"DROP TABLE {self.name}_old")

    def prepare_migration(self, db: sqlite3.Connection) -> List[Callable]:
        # migrate table by re-syncing entries from stored raw json replies
        try:
            db.execute(f"DROP TABLE IF EXISTS {self.name}_old")
            db.execute(f"ALTER TABLE {self.name} RENAME TO {self.name}_old")
            return [self.migrate]
        except sqlite3.DatabaseError:
            return []

    def upsert_row(self, db: sqlite3.Connection, row: Mapping[str, Any]) -> None:
        keys = ', '.join(row.keys())
        placeholders = ', '.join('?' for k in row.keys())
        db.execute(
            f"INSERT OR REPLACE INTO {self.name} ({keys}) VALUES ({placeholders})",
            tuple(row.values())
        )

    def upsert(
        self,
        db: sqlite3.Connection,
        data: Iterable[Mapping[str, Any]],
        incremental: bool = False
    ):
        rows = (self._from_dict(datum) for datum in data)
        with db:  # transaction
            db.execute("BEGIN")

            old_ids = set(r['id'] for r in db.execute(f"SELECT id FROM {self.name}"))
            seen = 0
            new = 0
            deleted = 0

            for row in rows:
                logging.debug(f"{self.name}: {row['id']} {'seen' if row['id'] in old_ids else 'new'}")

                if row['id'] in old_ids:
                    old_ids.discard(row['id'])

                    seen += 1
                    if incremental and seen > 10:
                        break
                else:
                    new += 1

                self.upsert_row(db, row)

            if not incremental:
                delete = ((i,) for i in old_ids)
                db.executemany(f"DELETE FROM {self.name} WHERE id = ?", delete)
                deleted += len(old_ids)

            logging.info(f"{self.name} upsert stats: {new} new, {seen} seen, {deleted} deleted")


@dataclass(frozen=True)
class Schema:
    version: int
    tables: List[Table]

    def initialize(self, db: sqlite3.Connection) -> None:
        with db:  # transaction
            db.execute("BEGIN")

            migrations = self.prepare_migrations(db)
            for table in self.tables:
                table.create(db)
            for migration in migrations:
                migration(db)

    def prepare_migrations(self, db: sqlite3.Connection) -> List[Callable]:
        db_version = db.execute("PRAGMA user_version").fetchone()[0]
        if db_version >= self.version:
            return []

        migrations: List[Callable] = []
        for table in self.tables:
            migrations += table.prepare_migration(db)
        migrations.append(lambda db: db.execute(f"PRAGMA user_version = {self.version}"))

        return migrations


@contextmanager
def database(path: Union[str, Path], schema: Schema) -> Iterator[sqlite3.Connection]:
    if isinstance(path, Path):
        path.parent.mkdir(parents=True, exist_ok=True)

    db = sqlite3.connect(path, isolation_level=None)
    db.row_factory = sqlite3.Row
    try:
        schema.initialize(db)
        yield db
    finally:
        db.close()
